from __future__ import annotations

from abc import abstractmethod, abstractproperty
from dataclasses import dataclass
from datetime import datetime, timedelta
from typing import overload, Optional, Set, Protocol


class TimeRange(Protocol):

    @abstractmethod
    def __contains__(self, moment: datetime) -> bool:
        """
        Check whether a moment is withing the TimeRange. Both endpoints are included.
        """

    @property
    @abstractmethod
    def to_timedelta(self) -> timedelta:
        pass

    # @abstractmethod
    # def __and__(self, other: TimeRange) -> Optional[TimeRange]:
    #     """
    #     Find the intersection of the two TimeRange's: A ⋂ B.
    #     """

    @abstractmethod
    def __add__(self, other: TimeRange) -> Optional[TimeRange]:
        """
        Find the union of the two TimeRange's: A ⋃ B.
        """


@dataclass(frozen=True)
class ContinuousTimeRange(TimeRange):
    start: datetime
    end: datetime

    def __post_init__(self):
        if self.start > self.end:
            raise ValueError("Start cannot be later than end.")

    def __contains__(self, moment: datetime) -> bool:
        return self.start <= moment <= self.end

    @property
    def to_timedelta(self) -> timedelta:
        return self.end - self.start

    def __and__(self, other: ContinuousTimeRange) -> Optional[ContinuousTimeRange]:
        if type(self) != type(other):
            return NotImplemented
        if not self.intersects_with(other):
            return None
        return ContinuousTimeRange(max(self.start, other.start), min(self.end, other.end))

    def __add__(self, other) -> Optional[ContinuousTimeRange]:
        if type(self) != type(other):
            return NotImplemented
        if not self.intersects_with(other):
            return None
        return ContinuousTimeRange(min(self.start, other.start), max(self.end, other.end))

    def intersects_with(self, other: ContinuousTimeRange) -> bool:
        # FIXME This does not work if `other` is a superset of `this`
        #  e.g. self=today, other=this_year
        return other.start in self or other.end in self


class DiscontinuousTimeRange(TimeRange):

    @overload
    def __init__(self):
        pass

    @overload
    def __init__(self, start: datetime, end: datetime):
        pass

    @overload
    def __init__(self, cls, start: datetime, duration: timedelta):
        pass

    def __init__(self, start: Optional[datetime] = None,
                 end: Optional[datetime] = None,
                 duration: Optional[timedelta] = None):
        self._periods: Set[ContinuousTimeRange] = set()
        if (start and not (end or duration)) or (end and duration):
            raise ValueError("Allowed combination of constructor parameters are:\n  "
                             "* Empty constructor\n  * `start` & `end`\n  * `start` & `duration`")
        if start:
            self._periods.add(ContinuousTimeRange(start, end or start + duration))

    def __repr__(self) -> str:
        return ' + '.join(
            [f"TimeRange(start={repr(p.start)}, end={repr(p.end)})" for p in self._periods]
        ) or "TimeRange()"

    def __bool__(self) -> bool:
        """
        A TimeRange evaluates to True if its duration is grater than zero.
        """
        return len(self._periods) != 0

    def __contains__(self, moment: datetime) -> bool:
        return any([moment in p for p in self._periods])

    def __add__(self, other) -> DiscontinuousTimeRange:
        if type(self) != type(other):
            return NotImplemented
        intersectionless_periods: Set[ContinuousTimeRange] = set()
        for p in {*self._periods, *other._periods}:
            for per in intersectionless_periods:
                if p.intersects_with(per):
                    intersectionless_periods.discard(per)
                    intersectionless_periods.add(p + per)
                    break
            else:
                intersectionless_periods.add(p)
        result = DiscontinuousTimeRange()
        result._periods = intersectionless_periods
        return result

    @property
    def to_timedelta(self) -> timedelta:
        return sum([p.to_timedelta for p in self._periods], start=timedelta())



import unittest
from datetime import datetime, timedelta

from timeset import DiscontinuousTimeRange, ContinuousTimeRange

date = (2021, 5, 20)
start = datetime(*date, 12, 12)
end = datetime(*date, 14, 12)
string_repr = "TimeRange(start=datetime.datetime(2021, 5, 20, 12, 12), end=datetime.datetime(2021, 5, 20, 14, 12))"


class ContinuousTimeRangeTest(unittest.TestCase):

    def setUp(self) -> None:
        self.timerange = ContinuousTimeRange(start, end)

    def test_initialization(self):
        with self.assertRaisesRegex(ValueError, "Start cannot be later than end."):
            ContinuousTimeRange(end, start)

    def test_contains(self):
        self.assertTrue(datetime(*date, 13, 12) in self.timerange)
        # Assert that endpoints are included
        self.assertTrue(start in self.timerange)
        self.assertTrue(end in self.timerange)

    def test_does_not_contain(self):
        self.assertFalse(datetime(*date, 22, 12) in self.timerange)

    def test_as_timedelta(self):
        self.assertEqual(ContinuousTimeRange(start, start).to_timedelta, timedelta(hours=0))
        self.assertEqual(self.timerange.to_timedelta, timedelta(hours=2))

    def test_intersection(self):
        union1 = self.timerange & ContinuousTimeRange(datetime(*date, 13, 12), datetime(*date, 22, 12))
        self.assertEqual(union1, ContinuousTimeRange(datetime(*date, 13, 12), end))

        union2 = self.timerange & ContinuousTimeRange(start, start)
        self.assertEqual(union2, ContinuousTimeRange(start, start))

    def test_union(self):
        union1 = self.timerange + ContinuousTimeRange(datetime(*date, 13, 12), datetime(*date, 15, 12))
        self.assertEqual(union1, ContinuousTimeRange(start, datetime(*date, 15, 12)))

        union2 = self.timerange + ContinuousTimeRange(start, start)
        self.assertEqual(union2, ContinuousTimeRange(start, end))

    def test_intersects_with(self):
        self.assertTrue(self.timerange.intersects_with(
            ContinuousTimeRange(datetime(*date, 13, 12), datetime(*date, 15, 12))))

    def test_does_not_intersect_with(self):
        self.assertFalse(self.timerange.intersects_with(
            ContinuousTimeRange(datetime(*date, 18, 12), datetime(*date, 19, 12))))


class TimeRangeInitializationTest(unittest.TestCase):
    def test_empty_initialization(self):
        t = DiscontinuousTimeRange()
        self.assertEqual(str(t), "TimeRange()")

    def test_initialization_with_start_and_end(self):
        t = DiscontinuousTimeRange(start=start, end=end)
        self.assertEqual(str(t), string_repr)

    def test_initialization_with_start_and_duration(self):
        t = DiscontinuousTimeRange(start=start, duration=timedelta(hours=8))
        self.assertEqual(t.to_timedelta, timedelta(hours=8))
        self.assertTrue(start in t)
        self.assertTrue(start + timedelta(hours=8) in t)

    def test_start_end_integrity(self):
        with self.assertRaises(ValueError):
            DiscontinuousTimeRange(start=end, end=start)


class TimeRangeBooleanTest(unittest.TestCase):
    def test_bool_true(self):
        t = DiscontinuousTimeRange(start=start, end=end)
        self.assertTrue(t)

    def test_bool_false(self):
        t = DiscontinuousTimeRange()
        self.assertFalse(t)


class TimeRangeTimedeltaTest(unittest.TestCase):
    def test_timedelta_two_hours(self):
        t = DiscontinuousTimeRange(start=start, end=end)
        self.assertEqual(t.to_timedelta, timedelta(hours=2))

    def test_timedelta_zero(self):
        t = DiscontinuousTimeRange()
        self.assertEqual(t.to_timedelta, timedelta(hours=0))


class TimeRangeContainsTest(unittest.TestCase):
    def test_contains(self):
        t = DiscontinuousTimeRange(start=start, end=end)
        self.assertTrue(datetime(2021, 5, 20, 12, 13) in t)

    def test_does_not_contain(self):
        t = DiscontinuousTimeRange(start=start, end=end)
        self.assertFalse(datetime(2021, 5, 20, 16, 12) in t)


class TimeRangeAdditionTest(unittest.TestCase):
    def test_no_overlap(self):
        t1 = DiscontinuousTimeRange(start=datetime(*date, 12), duration=timedelta(hours=2))
        t2 = DiscontinuousTimeRange(start=datetime(*date, 16), duration=timedelta(hours=2))
        self.assertEqual((t1 + t2).to_timedelta, timedelta(hours=4))

    def test_one_contains_another(self):
        t1 = DiscontinuousTimeRange(start=datetime(*date, 12), duration=timedelta(hours=2))
        t2 = DiscontinuousTimeRange(start=datetime(*date, 13), duration=timedelta(hours=1))
        self.assertEqual((t1 + t2).to_timedelta, timedelta(hours=2))

    def test_overlap(self):
        t1 = DiscontinuousTimeRange(start=datetime(*date, 12), duration=timedelta(hours=3))
        t2 = DiscontinuousTimeRange(start=datetime(*date, 13), duration=timedelta(hours=3))
        self.assertEqual((t1 + t2).to_timedelta, timedelta(hours=4))


if __name__ == '__main__':
    unittest.main()
